import manuel
import manuel.testing
import re
import string
import textwrap

punctuation = re.escape(string.punctuation)
SECTION_TITLE = re.compile(r'^.+$', re.MULTILINE)
SECTION_UNDERLINE = re.compile('^[' + punctuation + ']+\s*$', re.MULTILINE)
MARKER = re.compile(r'^.. test-case: (\S+)', re.MULTILINE)


def find_section_headers(document):
    for region in document.find_regions(SECTION_TITLE, SECTION_UNDERLINE):
        # regions that represent titles will have two lines
        if region.source.count('\n') != 2:
            continue

        title, underline = region.source.splitlines()

        # the underline has to be the same length as or longer than the title
        if len(underline) < len(title):
            continue

        # ok, this is a region we want
        document.claim_region(region)

        test_case_name = title.strip()
        region.parsed = manuel.testing.TestCaseMarker(test_case_name)


def find_markers(document):
    for region in document.find_regions(MARKER):
        document.claim_region(region)
        test_case_name = region.start_match.group(1)
        region.parsed = manuel.testing.TestCaseMarker(test_case_name)


class SectionManuel(manuel.Manuel):
    def __init__(self):
        manuel.Manuel.__init__(self, [find_section_headers])


class MarkerManuel(manuel.Manuel):
    def __init__(self):
        manuel.Manuel.__init__(self, [find_markers])



import doctest as real_doctest
import functools
import inspect
import io
import itertools
import manuel
import os.path
import re
import sys
import types
import unittest

__all__ = ['TestSuite', 'TestFactory']


class TestCaseMarker(object):
    def __init__(self, id=''):
        self.id = id


class TestCase(unittest.TestCase):
    def __init__(self, m, regions, globs, setUp=None, tearDown=None):
        unittest.TestCase.__init__(self)
        self.manuel = m
        self.regions = regions
        self.globs = globs
        self.setUp_func = setUp
        self.tearDown_func = tearDown

    def setUp(self):
        if self.setUp_func is not None:
            self.setUp_func(self)

    def tearDown(self):
        if self.tearDown_func is not None:
            self.tearDown_func(self)

    def runTest(self):
        self.regions.evaluate_with(self.manuel, self.globs)
        self.regions.format_with(self.manuel)
        results = [r.formatted for r in self.regions if r.formatted]
        if results:
            DIVIDER = '-' * 70 + '\n'
            raise real_doctest.DocTestCase.failureException(
                '\n' + DIVIDER + DIVIDER.join(results)
            )

    def debug(self):
        self.setUp()
        self.manuel.debug = True
        self.regions.evaluate_with(self.manuel, self.globs)
        self.tearDown()

    def countTestCases(self):
        return len([r for r in self.regions if r.parsed])

    def shortDescription(self):
        if self.regions.id:
            return self.regions.location + ':' + self.regions.id
        else:
            return self.regions.location

    __str__ = __repr__ = id = shortDescription


def group_regions_by_test_case(document):
    """Generate groups of regions according to which testcase they belong"""
    document_iter = iter(document)
    marker = None
    while True:
        accumulated_regions = manuel.RegionContainer()
        while True:
            region = None  # being defensive
            try:
                region = next(document_iter)
            except StopIteration:
                if not accumulated_regions:
                    break
            else:
                accumulated_regions.append(region)

                if not isinstance(region.parsed, TestCaseMarker):
                    continue

            # we just found a test case marker or hit the end of the
            # document

            # figure out what this test case's ID is
            accumulated_regions.location = document.location
            if marker is not None and marker.parsed.id:
                accumulated_regions.id = marker.parsed.id

            yield accumulated_regions
            marker = region
            break

        # if there are no more regions, stop
        try:
            region = next(document_iter)
        except StopIteration:
            break

        # put the region we peeked at back so the inner loop can consume it
        document_iter = itertools.chain([region], document_iter)


# copied from zope.testing.doctest
def _module_relative_path(module, path):
    if not inspect.ismodule(module):
        raise TypeError('Expected a module: %r' % module)
    if path.startswith('/'):
        raise ValueError('Module-relative files may not have absolute paths')

    # Find the base directory for the path.
    if hasattr(module, '__file__'):
        # A normal module/package
        basedir = os.path.split(module.__file__)[0]
    elif module.__name__ == '__main__':
        # An interactive session.
        if len(sys.argv) > 0 and sys.argv[0] != '':
            basedir = os.path.split(sys.argv[0])[0]
        else:
            basedir = os.curdir
    else:
        # A module w/o __file__ (this includes builtins)
        raise ValueError(
            "Can't resolve paths relative to the module "
            + module
            + " (it has no __file__)"
        )

    # Combine the base directory and the path.
    return os.path.join(basedir, *(path.split('/')))


def TestSuite(m, *paths, **kws):
    """A unittest suite that processes files with Manuel

    The path to each document file is given as a string.

    A number of options may be provided as keyword arguments:

    `setUp`
      A set-up function.  This is called before running the tests in each file.
      The setUp function will be passed a TestCase object.  The setUp function
      can access the test globals as the `globs` attribute of the instance
      passed.

    `tearDown`
      A tear-down function.  This is called after running the tests in each
      file.  The tearDown function will be passed a Manuel object.  The
      tearDown function can access the test globals as the `globs` attribute of
      the instance passed.

    `globs`
      A dictionary containing initial global variables for the tests.

    `TestCase`
      The TestCase class to be used instead of manuel.testing.TestCase.

    """

    suite = unittest.TestSuite()
    globs = kws.pop('globs', {})
    TestCase_class = kws.pop('TestCase', TestCase)

    # walk up the stack frame to find the module that called this function
    for depth in range(1, 5):
        try:
            calling_module = sys.modules[sys._getframe(depth).f_globals['__name__']]
        except KeyError:
            continue
        else:
            break

    for path in paths:
        if os.path.isabs(path):
            abs_path = os.path.normpath(path)
        else:
            abs_path = os.path.abspath(_module_relative_path(calling_module, path))

        with io.open(abs_path, 'rt', newline=None) as fp:
            contents = fp.read()
            if not isinstance(contents, str):
                # Python 2, we read unicode, but we really need a str
                contents = contents.encode("utf-8")
            document = manuel.Document(contents, location=abs_path)
        document.parse_with(m)

        for regions in group_regions_by_test_case(document):
            tc = TestCase_class(m, regions, globs, **kws)
            tc.globs['self'] = tc
            suite.addTest(tc)

    return suite


_not_word = re.compile(r'\W')


class TestFactory:
    def __init__(self, m):
        self.m = m

    def __call__(self, path):
        base = os.path.dirname(os.path.abspath(sys._getframe(2).f_globals['__file__']))
        path = os.path.join(base, path)
        with open(path) as f:
            test = f.read()

        m = self.m

        def test_file(self, setup=lambda i: None):
            if isinstance(self, types.FunctionType):
                # We're being used as a decorator. `self` is a setup method.
                f = functools.wraps(self)(lambda inst: test_file(inst, self))
                f.filepath = path
                f.filename = os.path.basename(path)
                return f

            setup(self)
            globs = dict(getattr(self, 'globs', ()))
            globs['test'] = self
            document = manuel.Document(test, location=path)
            document.parse_with(m)
            [regions] = group_regions_by_test_case(document)
            TestCase(m, regions, globs).runTest()

        test_file.filepath = path
        test_file.filename = filename = os.path.basename(path)
        name = _not_word.sub('_', os.path.splitext(filename)[0])
        if not name.startswith('test'):
            name = 'test_' + name

        test_file.__name__ = name

        return test_file



import manuel
import re
import string
import textwrap

CAPTURE_DIRECTIVE = re.compile(
    r'^(?P<indent>(\t| )*)\.\.\s*->\s*(?P<name>\S+).*$', re.MULTILINE
)


class Capture(object):
    def __init__(self, name, block):
        self.name = name
        self.block = block


def normalize_whitespace(s):
    return s.replace('\t', ' ' * 8)  # turn tabs into spaces


@manuel.timing(manuel.EARLY)
def find_captures(document):
    while True:
        regions = document.find_regions(CAPTURE_DIRECTIVE)
        if not regions:
            break
        region = regions[-1]
        # note that start and end have different bases, "start" is the offset
        # from the begining of the region, "end" is a document line number
        end = region.lineno - 2

        indent = region.start_match.group('indent')
        indent = normalize_whitespace(indent)

        def indent_matches(line):
            """Is the indentation of a line match what we're looking for?"""
            line = normalize_whitespace(line)

            if not line.strip():
                # the line consists entirely of whitespace (or nothing at all),
                # so is not considered to be of the appropriate indentation
                return False

            if line.startswith(indent):
                if line[len(indent)] not in string.whitespace:
                    return True

            # if none of the above found the indentation to be a match, it is
            # not a match
            return False

        # now that we've extracted the information we need, lets slice up the
        # document's regions to match

        for candidate in document:
            if candidate.lineno >= region.lineno:
                break
            found_region = candidate

        lines = found_region.source.splitlines()
        if found_region.lineno + len(lines) < end:
            raise RuntimeError('both start and end lines must be in the ' 'same region')

        start = None
        for offset, line in reversed(list(enumerate(lines))):
            if offset > end - found_region.lineno:
                continue
            if indent_matches(line):
                break
            start = offset + 1

        if start is None:
            raise RuntimeError(
                "couldn't find the start of the block; "
                "improper indentation of capture directive?"
            )

        _, temp_region = document.split_region(
            found_region, found_region.lineno + start
        )

        # there are some extra lines in the new region, trim them off
        final_region, _ = document.split_region(temp_region, end + 1)
        document.remove_region(final_region)

        name = region.start_match.group('name')
        block = textwrap.dedent(final_region.source)
        document.claim_region(region)
        region.parsed = Capture(name, block)


def store_capture(region, document, globs):
    if not isinstance(region.parsed, Capture):
        return

    globs[region.parsed.name] = region.parsed.block


class Manuel(manuel.Manuel):
    def __init__(self):
        manuel.Manuel.__init__(self, [find_captures], [store_capture])



import re

# constants for use with "timing" decorator
EARLY = 'early'
LATE = 'late'


def timing(timing):
    assert timing in (EARLY, LATE)

    def decorate(func):
        func.manuel_timing = timing
        return func

    return decorate


def newlineify(s):
    if s == '' or s[-1] != '\n':
        s += '\n'
    return s


class Region(object):
    """A portion of source found via regular expression."""

    parsed = None
    evaluated = None
    formatted = None

    def __init__(
        self, lineno, source, start_match=None, end_match=None, provenance=None
    ):
        self.lineno = lineno
        self.source = newlineify(source)
        self.start_match = start_match
        self.end_match = end_match
        self.provenance = provenance

    def copy(self):
        """Private utility function to make a copy of this region."""
        copy = Region(self.lineno, self.source, provenance=self.provenance)
        copy.parsed = self.parsed
        copy.evaluated = self.evaluated
        copy.formatted = self.formatted
        return copy


def find_line(region, index):
    return region[:index].count('\n') + 1


def check_region_start(region, match):
    if match.start() != 0 and region.source[match.start() - 1] != '\n':
        raise ValueError('Regions must start at the begining of a line.')


def check_region_end(region, match):
    if match.end() != len(region.source) and region.source[match.end() - 1] != '\n':
        raise ValueError('Regions must end at the ending of a line.')


def lines_to_string(lines):
    return '\n'.join(lines) + '\n'


def make_string_into_lines(s):
    lines = newlineify(s).split('\n')
    assert lines[-1] == ''
    del lines[-1]
    return lines


def break_up_region(original, new):
    assert original.parsed is None
    lines = make_string_into_lines(original.source)

    new_regions = []

    # figure out if there are any lines before the given region
    before_lines = lines[: new.lineno - original.lineno]
    if before_lines:
        new_regions.append(Region(original.lineno, lines_to_string(before_lines)))

    # put in the parsed
    new_regions.append(new)

    # figure out if there are any lines after the given region
    assert new.source[-1] == '\n', 'all lines must end with a newline'
    lines_in_new = new.source.count('\n')
    after_lines = lines[len(before_lines) + lines_in_new :]
    if after_lines:
        first_line_after_new = new.lineno + lines_in_new
        new_regions.append(Region(first_line_after_new, lines_to_string(after_lines)))

    assert original.source.count('\n') == sum(r.source.count('\n') for r in new_regions)
    return new_regions


def sort_handlers(handlers):
    def key(f):
        # "j" was chosen because it sorts between "early" and "late"
        return getattr(f, 'manuel_timing', 'j')

    return sorted(handlers, key=key)


def find_end_of_line(s):
    end = 0
    while len(s) < end and s[end] != '\n':
        end += 1
    return end


class RegionContainer(object):

    location = '<memory>'
    id = None

    def __init__(self):
        self.regions = []

    def parse_with(self, m):
        for parser in sort_handlers(m.parsers):
            parser(self)

    def evaluate_with(self, m, globs):
        for region in list(self):
            for evaluater in sort_handlers(m.evaluaters):
                evaluater(region, self, globs)

    def format_with(self, m):
        for formatter in sort_handlers(m.formatters):
            formatter(self)

    def process_with(self, m, globs):
        """Run all phases of document processing using a Manuel instance."""
        self.parse_with(m)
        self.evaluate_with(m, globs)
        self.format_with(m)

    def formatted(self):
        """Return a string of all non-boolean-false formatted regions."""
        return ''.join(region.formatted for region in self if region.formatted)

    def append(self, region):
        self.regions.append(region)

    def __iter__(self):
        """Iterate over all regions of the document."""
        return iter(self.regions)

    def __bool__(self):
        return bool(self.regions)


class Document(RegionContainer):
    def __init__(self, source, location=None):
        RegionContainer.__init__(self)

        if location is not None:
            self.location = location

        self.source = newlineify(source)

        self.append(Region(lineno=1, source=source))
        self.shadow_regions = []

    def find_regions(self, start, end=None):
        def compile(regex):
            if regex is not None and isinstance(regex, str):
                regex = re.compile(regex)
            return regex

        start = compile(start)
        end = compile(end)

        results = []
        for region in self.regions:
            # can't parse things that have already been parsed
            if region.parsed:
                continue

            for start_match in re.finditer(start, region.source):
                first_lineno = (
                    region.lineno + find_line(region.source, start_match.start()) - 1
                )
                check_region_start(region, start_match)

                if end is None:
                    end_match = None

                    text = start_match.group()
                else:
                    end_match = end.search(region.source, start_match.end())

                    # couldn't find a match for the end re, try again
                    if end_match is None:
                        continue
                    end_position = end_match.end() + find_end_of_line(
                        region.source[end_match.end() :]
                    )
                    text = region.source[start_match.start() : end_position]

                if text[-1] != '\n':
                    text += '\n'

                new_region = Region(first_lineno, text, start_match, end_match)
                self.shadow_regions.append(new_region)
                results.append(new_region)

        return results

    def split_region(self, region, lineno):
        lineno -= region.lineno
        assert lineno > 0
        assert region in self.regions
        assert region.parsed == region.evaluated == region.formatted == None
        lines = make_string_into_lines(region.source)
        source1 = lines_to_string(lines[:lineno])
        source2 = lines_to_string(lines[lineno:])
        region_index = self.regions.index(region)
        del self.regions[region_index]
        lines_in_source1 = source1.count('\n')
        region1 = Region(region.lineno, source1)
        region2 = Region(region.lineno + lines_in_source1, source2)
        self.regions.insert(region_index, region2)
        self.regions.insert(region_index, region1)
        if not region.source == source1 + source2:
            raise RuntimeError(
                'when splitting a region, combined results do ' 'not equal the input'
            )
        return region1, region2

    def claim_region(self, to_be_replaced):
        new_regions = []
        old_regions = list(self.regions)
        while old_regions:
            region = old_regions.pop(0)
            if region.lineno == to_be_replaced.lineno:
                assert not region.parsed
                new_regions.extend(break_up_region(region, to_be_replaced))
                break
            elif region.lineno > to_be_replaced.lineno:  # we "overshot"
                assert not new_regions[-1].parsed
                to_be_broken = new_regions[-1]
                del new_regions[-1]
                new_regions.extend(break_up_region(to_be_broken, to_be_replaced))
                new_regions.append(region)
                break

            new_regions.append(region)
        else:
            # we didn't make any replacements, so the parsed data must be for
            # the very last region, which also must not have been parsed yet
            assert not region.parsed
            del new_regions[-1]
            new_regions.extend(break_up_region(region, to_be_replaced))

        new_regions.extend(old_regions)
        self.regions = new_regions

    def insert_region(self, where, marker_region, new_region):
        if new_region in self.regions:
            raise ValueError(
                'Only regions not already in the document may be inserted.'
            )
        if new_region in self.shadow_regions:
            raise ValueError(
                'Regions returned by "find_regions" can not be directly '
                'inserted into a document.  Use "claim_region" instead.'
            )

        for index, region in enumerate(self.regions):
            if region is marker_region:
                if where == 'after':
                    index += 1
                self.regions.insert(index, new_region)
                break

    def remove_region(self, region):
        self.regions.remove(region)

    def insert_region_before(self, marker_region, new_region):
        self.insert_region('before', marker_region, new_region)

    def insert_region_after(self, marker_region, new_region):
        self.insert_region('after', marker_region, new_region)


def call(func):
    return func()


class Manuel(object):

    _debug = False

    def __init__(self, parsers=None, evaluaters=None, formatters=None):
        if parsers is not None:
            self.parsers = parsers
        else:
            self.parsers = []

        if evaluaters is not None:
            self.evaluaters = evaluaters
        else:
            self.evaluaters = []

        if formatters is not None:
            self.formatters = formatters
        else:
            self.formatters = []

        # other instances that this one has been extended with
        self.others = []

    def add_parser(self, parser):
        self.parsers.append(parser)

    def add_evaluater(self, evaluater):
        self.evaluaters.append(evaluater)

    def add_formatter(self, formatter):
        self.formatters.append(formatter)

    def __extend(self, other):
        self.others.append(other)
        self.debug = max(self.debug, other.debug)
        self.parsers.extend(other.parsers)
        self.evaluaters.extend(other.evaluaters)
        self.formatters.extend(other.formatters)

    # the testing integration (manuel.testing) sets this flag when needed
    @call
    def debug():
        def getter(self):
            debug = self._debug
            if self.others:
                debug = max(debug, max(m.debug for m in self.others))
            return debug

        def setter(self, value):
            self._debug = value
            for m in self.others:
                m.debug = value

        return property(getter, setter)

    def __add__(self, other):
        m = Manuel()
        m.__extend(self)
        m.__extend(other)
        return m



import manuel
import re

FOOTNOTE_REFERENCE_LINE_RE = re.compile(r'^.*\[([^\]]+)]_.*$', re.MULTILINE)
FOOTNOTE_REFERENCE_RE = re.compile(r'\[([^\]]+)]_')
FOOTNOTE_DEFINITION_RE = re.compile(r'^\.\.\s*\[\s*([^\]]+)\s*\].*$', re.MULTILINE)
END_OF_FOOTNOTE_RE = re.compile(r'^\S.*$', re.MULTILINE)


class FootnoteReference(object):
    def __init__(self, names):
        self.names = names


class FootnoteDefinition(object):
    def __init__(self, name):
        self.name = name


@manuel.timing(manuel.EARLY)
def find_footnote_references(document):
    # find the markers that show where footnotes have been defined.
    footnote_names = []
    for region in document.find_regions(FOOTNOTE_DEFINITION_RE):
        name = region.start_match.group(1)
        document.claim_region(region)
        region.parsed = FootnoteDefinition(name)
        footnote_names.append(name)

    # find the markers that show where footnotes have been referenced.
    for region in document.find_regions(FOOTNOTE_REFERENCE_LINE_RE):
        assert region.source.count('\n') == 1
        names = FOOTNOTE_REFERENCE_RE.findall(region.source)
        for name in names:
            if name not in footnote_names:
                raise RuntimeError('Unknown footnote: %r' % name)

        assert names
        document.claim_region(region)
        region.parsed = FootnoteReference(names)


@manuel.timing(manuel.LATE)
def do_footnotes(document):
    """Copy footnoted items into their appropriate position."""
    # first find all the regions that are in footnotes
    footnotes = {}
    name = None
    for region in list(document):
        if isinstance(region.parsed, FootnoteDefinition):
            name = region.parsed.name
            footnotes[name] = []
            document.remove_region(region)
            continue

        if END_OF_FOOTNOTE_RE.search(region.source):
            name = None

        if name is not None:
            footnotes[name].append(region)
            document.remove_region(region)

    # now make copies of the footnotes in the right places
    for region in list(document):
        if not isinstance(region.parsed, FootnoteReference):
            continue
        names = region.parsed.names
        for name in names:
            for footnoted in footnotes[name]:
                document.insert_region_before(region, footnoted.copy())
        document.remove_region(region)


class Manuel(manuel.Manuel):
    def __init__(self):
        manuel.Manuel.__init__(self, [find_footnote_references, do_footnotes])



import manuel
import re
import textwrap

RESET = re.compile(r'^\.\.\s*reset-globs\s*$', re.MULTILINE)
CAPTURE = re.compile(r'^\.\.\s*capture-globs\s*$', re.MULTILINE)

baseline = {}


class Reset(object):
    pass


def find_reset(document):
    for region in document.find_regions(RESET):
        document.claim_region(region)
        region.parsed = Reset()


def execute_reset(region, document, globs):
    if not isinstance(region.parsed, Reset):
        return

    globs.clear()
    globs.update(baseline)


class Capture(object):
    pass


def find_baseline(document):
    # clear the baseline globs at the begining of a run (a bit of a hack)
    baseline.clear()

    for region in document.find_regions(CAPTURE):
        document.claim_region(region)
        region.parsed = Capture()


def execute_baseline(region, document, globs):
    if not isinstance(region.parsed, Capture):
        return

    baseline.clear()
    baseline.update(globs)


class Manuel(manuel.Manuel):
    def __init__(self):
        manuel.Manuel.__init__(
            self, [find_reset, find_baseline], [execute_reset, execute_baseline]
        )



import doctest
import manuel
import os.path
from io import StringIO

DocTestRunner = doctest.DocTestRunner
DebugRunner = doctest.DebugRunner


class DocTestResult(StringIO):
    pass


def parse(m, document, parser):
    for region in list(document):
        if region.parsed:
            continue
        region_start = region.lineno
        region_end = region.lineno + region.source.count('\n')
        for chunk in parser.parse(region.source):
            # If the chunk contains prose (as opposed to and example), skip it.
            if isinstance(chunk, str):
                continue

            chunk._manual = m
            chunk_line_count = chunk.source.count('\n') + chunk.want.count('\n')

            split_line_1 = region_start + chunk.lineno
            split_line_2 = split_line_1 + chunk_line_count

            # if there is some source we need to trim off the front...
            if split_line_1 > region.lineno:
                _, region = document.split_region(region, split_line_1)

            if split_line_2 < region_end:
                found, region = document.split_region(region, split_line_2)
            else:
                found = region

            document.claim_region(found)

            # Since we're treating each example as a stand-alone thing, we need
            # to reset its line number to zero.
            chunk.lineno = 0
            found.parsed = chunk

            assert region in document


class DocTest(doctest.DocTest):
    def __init__(self, examples, globs, name, filename, lineno, docstring):
        # do everything like regular doctests, but don't make a copy of globs
        doctest.DocTest.__init__(
            self, examples, globs, name, filename, lineno, docstring
        )
        self.globs = globs


def evaluate(m, region, document, globs):
    # If the parsed object is not a doctest Example then we don't need to
    # handle it.

    if getattr(region.parsed, '_manual', None) is not m:
        return

    result = DocTestResult()
    test_name = os.path.split(document.location)[1]
    if m.debug:
        runner = m.debug_runner
        out = None
    else:
        runner = m.runner
        out = result.write

    # Use the testrunner-set option flags when running these tests.
    old_optionflags = runner.optionflags
    runner.optionflags |= doctest._unittest_reportflags
    runner.DIVIDER = ''  # disable unwanted result formatting

    # Here's where everything happens.
    example = region.parsed
    runner.run(
        DocTest(
            [example], globs, test_name, document.location, region.lineno - 1, None
        ),
        out=out,
        clear_globs=False,
    )

    runner.optionflags = old_optionflags  # Reset the option flags.
    region.evaluated = result


def format(document):
    for region in document:
        if not isinstance(region.evaluated, DocTestResult):
            continue
        region.formatted = region.evaluated.getvalue().lstrip()


class Manuel(manuel.Manuel):
    def __init__(self, optionflags=0, checker=None, parser=None):
        self.runner = DocTestRunner(
            optionflags=optionflags, checker=checker, verbose=False
        )
        self.debug_runner = DebugRunner(optionflags=optionflags, verbose=False)

        def evaluate_closure(region, document, globs):
            # capture "self"
            evaluate(self, region, document, globs)

        parser = parser or doctest.DocTestParser()
        manuel.Manuel.__init__(
            self,
            [lambda document: parse(self, document, parser)],
            [evaluate_closure],
            [format],
        )



import re
import textwrap
from manuel import Manuel as BaseManuel
from manuel.codeblock import CodeBlock, execute_code_block

CODEBLOCK_START = re.compile(
    r"((^```python)|(^% invisible-code-block:\s+python)$)",
    re.MULTILINE,
)
CODEBLOCK_END = re.compile(r"(\n(?=```\n))|((?:% [\S ]*)\n(?=\n))")


def find_code_blocks(document):
    for region in document.find_regions(CODEBLOCK_START, CODEBLOCK_END):
        start_end = CODEBLOCK_START.search(region.source).end()
        source = textwrap.dedent(region.source[start_end:])
        # MyST comments
        source = re.sub(r'\n%[ ]?', '\n', source)
        source_location = "%s:%d" % (document.location, region.lineno)
        code = compile(source, source_location, "exec", 0, True)
        document.claim_region(region)
        region.parsed = CodeBlock(code, source)


class Manuel(BaseManuel):
    def __init__(self):
        BaseManuel.__init__(self, [find_code_blocks], [execute_code_block])



import manuel
import re
import textwrap

IGNORE_START = re.compile(r'^\.\.\s*ignore-next-block\s*$', re.MULTILINE)
IGNORE_END = re.compile(r'(?<!ignore-next-block)\n\n(?=\S)|\Z')

baseline = {}


def find_ignores(document):
    for region in document.find_regions(IGNORE_START, IGNORE_END):
        document.claim_region(region)
        region.parsed = object()
        document.remove_region(region)


class Manuel(manuel.Manuel):
    def __init__(self):
        manuel.Manuel.__init__(self, [find_ignores])



import manuel
import re
import textwrap

CODEBLOCK_START = re.compile(
    r'(^\.\.\s*(invisible-)?code(-block)?::?\s*python\b(?:\s*\:[\w-]+\:.*\n)*)',  # noqa
    re.MULTILINE,
)
CODEBLOCK_END = re.compile(r'(\n\Z|\n(?=\S))')


class CodeBlock(object):
    def __init__(self, code, source):
        self.code = code
        self.source = source


def find_code_blocks(document):
    for region in document.find_regions(CODEBLOCK_START, CODEBLOCK_END):
        start_end = CODEBLOCK_START.search(region.source).end()
        source = textwrap.dedent(region.source[start_end:])
        source_location = '%s:%d' % (document.location, region.lineno)
        code = compile(source, source_location, 'exec', 0, True)
        document.claim_region(region)
        region.parsed = CodeBlock(code, source)


def execute_code_block(region, document, globs):
    if not isinstance(region.parsed, CodeBlock):
        return

    exec(region.parsed.code, globs)
    del globs['__builtins__']  # exec adds __builtins__, we don't want it


class Manuel(manuel.Manuel):
    def __init__(self):
        manuel.Manuel.__init__(self, [find_code_blocks], [execute_code_block])
